#ifndef _ENERGYMANAGER_H
#define ENERGYMANAGER_H
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
typedef struct energymanager_t
{
	unsigned char module_instance; /* 8-bits to hold the module instance in question */
	unsigned char* next_payload; /* pointer to data that must be sent to IMM to continue */
	void (*next_response_handler)(); /* pointer to a function to handle the payload */
	unsigned char* snapshot_buffer; /* pointer to accumulated snapshot data for analysis */
	int errorcode; /* set when an unexpected state has occurred */
	char* data_label;
	unsigned char* data;
	char* data_units;
	int data_length;
	int ear_ac_precision; /* precision of EARac values, in tenths of a percent */
	int ear_dc_precision; /* precinios of EARdc */
	int pcap_enabled; /* whether pcap is currently enabled or not */
	int snapshot_required_fetches; /* number of fetches required to pull all histogram data in */
	int snapshot_buffer_size; /* total size of snapshot buffer */
	int snapshot_buffer_chunk; /* size of data that can be fetched at a time from snapshot buffer */
	int snapshot_remainder_size; /* size of the last request after the full requests have been done */
	int dc_histogram_smallest_bin; /* smallest wattage in the histogram in watts */
	int dc_histogram_bincount; /* number of bins (technically only 8 bits required)*/
	int dc_histogram_binrange; /* the range of watts each bin represents */
	int dc_histogram_binsize; /* bytes per bin reading */
	int dc_buffer_offset; /* bytes into the snapshot buffer we currently are */
	int dc_fetch_bufconfig; /* whether to follow up histogram config with a followup snapshot buffer */
	int dc_fetch_snapshotbuf; /* whether to follow up a snapshot config answer with actual data retrieval */
	int firstbuffer; /* flag to indicate whether any fetches have been done against the snapshot buffer yet */
	int current_buffer_tag; /* store the currently valid buffer tag to detect mid-retrieval data invalidation */
	int snapshot_total_fetches; /* remember the total number of fetches in case buffer invalidation happens during conversation */
	unsigned int *dc_histogram_counts; /* relative histogram counts */
	double dc_histogram_accuracy; /* percent accuracy of dc histogram data */
} EnergyManager;
#define ERROR_INVALID_RESPONSE 1
#define ERROR_UNSUPPORTED_REVISION 2
#define ERROR_UNSUPPORTED_FEATURE 3

EnergyManager *new_energymanager();
void init_energymanager(EnergyManager *em);
void prep_get_powerstatus(EnergyManager* em);
void parse_status_data(EnergyManager *em, unsigned char* data, int length);
void prep_get_precision(EnergyManager* em);
void prep_get_ac_energy(EnergyManager* em);
void prep_get_dc_energy(EnergyManager* em);
void prep_set_capenable(EnergyManager* em, int enable);
void prep_set_cap(EnergyManager* em, int cap);
void prep_get_mincap(EnergyManager* em);
void prep_get_cap(EnergyManager* em);
void prep_get_maxcap(EnergyManager* em);
void parse_init_data(EnergyManager *em, unsigned char* data, int length);
void parse_snapbuffer(EnergyManager *em, unsigned char* data, int length);
void parse_capping_data(EnergyManager * em, unsigned char* data, int length);
void parse_ear_data(EnergyManager * em, unsigned char* data, int length);
void parse_ear_cfgdata(EnergyManager * em, unsigned char* data, int length);
void parse_buffer_config(EnergyManager* em, unsigned char* data, int length);
void parse_histogram_config(EnergyManager* em, unsigned char* data, int length);
void prep_get_dc_histogram_config(EnergyManager * em);



#endif
