#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"
#include "../../EnergyManager.h"

typedef EnergyManager *IBM__EnergyManager;

MODULE = IBM::EnergyManager		PACKAGE = IBM::EnergyManager		

PROTOTYPES: ENABLE

IBM::EnergyManager
new(classname)
    char *classname
    CODE:
        RETVAL = new_energymanager();
	init_energymanager(RETVAL);
    OUTPUT:
        RETVAL

void
prep_retrieve_histogram(em)
	IBM::EnergyManager em
	CODE:
		em->dc_fetch_bufconfig=1; /* the module needs to know to chain requests */
		em->dc_fetch_snapshotbuf=1; /* therefore, we request chaining all the histogram functions */
		prep_get_dc_histogram_config(em);

void
prep_set_cap(em,cap)
	IBM::EnergyManager em
	int cap

void
prep_set_capenable(em,enable)
	IBM::EnergyManager em
	int enable

void
prep_get_powerstatus(em)
	IBM::EnergyManager em

void
prep_get_cap(em)
	IBM::EnergyManager em

void
prep_get_mincap(em)
	IBM::EnergyManager em

void
prep_get_maxcap(em)
	IBM::EnergyManager em

void
prep_get_precision(em)
	IBM::EnergyManager em

void
prep_get_ac_energy(em)
	IBM::EnergyManager em
	
void
prep_get_dc_energy(em)
	IBM::EnergyManager em
	
void
get_next_payload(em)
	IBM::EnergyManager em
	PREINIT:
		int size;
		int i;
	PPCODE:
		if (em->next_payload == 0) { XSRETURN_EMPTY; }
		size = em->next_payload[0];
		if (size == 0) {
			XSRETURN(0);
		}
		for (i=1;i<=size;i++) {
			XPUSHs(sv_2mortal(newSViv(em->next_payload[i])));
		}
		
int
has_more_payload(em)
	IBM::EnergyManager em
	PPCODE:
		XPUSHs(sv_2mortal(newSViv((long)em->next_payload)));

int
energy_dc_precision(em)
	IBM::EnergyManager em
	CODE:
		RETVAL=em->ear_dc_precision;
	OUTPUT:
		RETVAL
		
int
capping_enabled(em)
	IBM::EnergyManager em
	CODE:
		RETVAL=em->pcap_enabled;
	OUTPUT:
		RETVAL

int
energy_ac_precision(em)
	IBM::EnergyManager em
	CODE:
		RETVAL=em->ear_ac_precision;
	OUTPUT:
		RETVAL


void
extract_relative_histogram(em)
	IBM::EnergyManager em
	PREINIT:
		char * curkey;
		int cur;
		int i;
		HV * histogram;
	PPCODE:
		histogram = (HV*)sv_2mortal((SV *)newHV());
		curkey=(char*)malloc(32);
		cur=em->dc_histogram_smallest_bin;
		for (i=0; i < em->dc_histogram_bincount; i++) {
			cur+=em->dc_histogram_binrange;
            if (em->dc_histogram_counts[i] == 0) {
                continue;
            }
			snprintf(curkey,32,"%dw-%dw",cur,cur+em->dc_histogram_binrange-1);
			hv_store(histogram,curkey,strlen(curkey),newSVuv(em->dc_histogram_counts[i]),0);
		}
		XPUSHs(sv_2mortal(newRV((SV*)histogram)));
		XPUSHs(sv_2mortal(newSVnv(em->dc_histogram_accuracy)));
		

void
extract_data(em)
	IBM::EnergyManager em
	PREINIT:
		int i;
	PPCODE:
		if (!em->data || !em->data_units || !em->data_label || !em->data_length) {
			return;
		}
		XPUSHs(sv_2mortal(newSVpv(em->data_label,0)));
		XPUSHs(sv_2mortal(newSVpv(em->data_units,0)));
		for (i=0;i<em->data_length;i++) {
			XPUSHs(sv_2mortal(newSViv(em->data[i])));
		}

void
handle_next_payload(em, ...)
	IBM::EnergyManager em
	PREINIT:
		int size;
		int i;
		char* response_data;
	PPCODE:
		if (em->next_response_handler == 0) { 
			if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
				free(em->next_payload);
				em->next_payload=0;
			}
			return;
		}
		size = items-1;
		response_data=(char*)malloc(size);
		for (i=1;i<=size;i++) {
			if (!SvIOK(ST(i))) {
				free(response_data);
				response_data=0;
				return;
			}
			response_data[i-1]=(char)SvIV(ST(i));
		}
		em->next_response_handler(em,response_data,size);
		free(response_data);
		response_data=0;

void
print(em)
    IBM::EnergyManager em
    CODE:
        printf("%d\n",em->errorcode);
