;(function(){
    var nodelist;
    var cputempchart;
    var powerchart;
    var active = true;
    var currpower = [];
    var currpowernames = [];
    var currtemps = [];
    var currtempnames = [];
    var mintemp = null;
    var maxtemp = null;
    var minpower = null;
    var maxpower = null;
    var pendingpowerupdate = null;
    var pendingtempupdate = null;

    function pad(number) {
        if (number < 10) {
            return '0' + number;
        }
        return number;
    }

    function initpowerchart(pane) {
        nodelist = [];
        var node = getNodeName();  // is this sufficient?
        cputempchart = new_chart();
        powerchart = new_chart();
        pane.appendChild(cputempchart);
        pane.appendChild(powerchart);

        getRequest('/confluent-api/noderange/' + node + '/nodes/', updatenodelist);
    }

    function updatenodelist(data) {
        if (data.item) {
            var newnode = data.item.href.slice(0, -1);
            nodelist.push(newnode);
        }
        if ('_requestdone' in data) {
            updatepowerchart();
            updatecputempchart();
        }
    }

    function new_chart() {
        var thechart = document.createElement('div');
        thechart.style.display = 'inline-block';
        thechart.style.width = (0.49 * window.innerWidth) + 'px';
        thechart.style.height = (0.8 * window.innerHeight) + 'px';
        thechart.data = null;
        return thechart;
    }

    function updatepowerchart() {
        if (!active) {
            return;
        }
        currpower = [];
        currpowernames = [];
        getRequest('/confluent-api/nodes/' + getNodeName() + '/sensors/hardware/power/all', parsePower);
    }

    function updatecputempchart() {
        if (!active) {
            return;
        }
        // Clear data from last time
        currtempnames = [];
        currtemps = [];
        getRequest('/confluent-api/nodes/' + getNodeName() + '/sensors/hardware/temperature/all', parseTemp);
    }

    function parsePower(data){

        if(data.sensors != null){
            var sensors = data.sensors;
            sensors.forEach(function(sensor){
                var sensorName = sensor.name;
                var sensorValue = sensor.value;
                if(sensorValue != null){
                    addPowerToChart(sensorName, sensorValue);
                }
            });
        }

        // Update plot
        if ('_requestdone' in data){
            updateplot(powerchart, 'Node Power Consumption', 'Power (W)', minpower, maxpower, currpower, currpowernames, nodelist);
            pendingpowerupdate = setTimeout(updatepowerchart, 5000);
        }
    }

    function parseTemp(data){

        if(data.sensors != null){
            var sensors = data.sensors;
            sensors.forEach(function(sensor){
                var sensorName = sensor.name;
                var sensorValue = sensor.value;
                if(sensorValue != null){
                    addTempToChart(sensorName, sensorValue);
                }
            });
        }

        // Update plot
        if ('_requestdone' in data){
            updateplot(cputempchart, 'Temperature', 'Degrees (C)', mintemp, maxtemp, currtemps, currtempnames, nodelist);
            pendingtempupdate = setTimeout(updatecputempchart, 5000);
        }
    }

    function addPowerToChart(currentName, currentValue){

        // Add current power & name to lists
        if(!currpowernames.includes(currentName)){
            currpowernames.push(currentName);
            currpower.push(currentValue);
        }

        // Update min and max power
        if (minpower == null){
            minpower = currentValue;
        } else if (currentValue < minpower){
            minpower = currentValue;
        }
        if (maxpower == null){
            maxpower = currentValue;
        } else if (currentValue > maxpower){
            maxpower = currentValue;
        }
    }

    function addTempToChart(currentName, currentValue){

        // Add current temp & name to lists
        if(!currtempnames.includes(currentName)){
            currtempnames.push(currentName);
            currtemps.push(currentValue);
        }        

        // Update min and max temps
        if (mintemp == null){
            mintemp = currentValue;
        } else if (currentValue < mintemp){
            mintemp = currentValue;
        }
        if (maxtemp == null){
            maxtemp = currentValue;
        } else if (currentValue > maxtemp){
            maxtemp = currentValue;
        }
    }


    function updateplot(plot, title, metric, min, max, data, tracenames, elemlist){
        var thedata;
        var layout;

        // Get existing data, or create it if it doesn't exist
        if (plot.data == null){

            // set up thedata
            thedata = [];
            for (var i = 0; i < data.length; i++){ // i represents the sensor
                thedata.push({
                    y: [],
                    name: tracenames[i],  // name of sensor
                    showscale: true,
                    opacity: 0.99,
                    x: [],
                    type: 'scatter'
                });
            }

            // set up layout
            layout = {
                title: title,
                xaxis: {
                    title: 'Time',
                    titlefont: {
                        family: 'Arial, sans-serif',
                        size: 14,
                        color: 'black'
                    },
                    type: 'category'
                },
                yaxis: {
                    title: metric,
                    titlefont: {
                        family: 'Arial, sans-serif',
                        size: 14,
                        color: 'black'
                    },
                    type: 'linear'
                },
                showlegend: true
            };
        } else {
            thedata = plot.data;
            layout = plot.layout;
        }

        // Set min and max
        thedata[0].cmin = min;
        thedata[0].cmax = max;

        // Get time information
        var now = new Date();
        var currentTime = now.getHours() + ":" + pad(now.getMinutes()) + ":" + pad(now.getSeconds());

        // Add new information
        var nextData;
        for (var i = 0; i < data.length; i++){  // i represents the sensor
            nextData = data[i];  // the sensor's reading
            if(thedata[i].x.length >= 60){  // limit data points to 60
                thedata[i].y.shift();
                thedata[i].x.shift();
            }
            thedata[i].y.push(nextData);
            thedata[i].x.push(currentTime);
        }

        // Plot
        if (plot.data == null) {
            Plotly.newPlot(plot, thedata, layout);
        }else{
            Plotly.redraw(plot);
        }
    }

    window.initpowerchart = initpowerchart;

})();